#include "RayTracer.h"

#define RAY_EPSILON 0.0001

#ifndef M_PI
#define M_PI 3.1415926535897932384626433832795
#endif

namespace RayTracer {
	bool gDebug;
	
	Image render(Vertex<3> cameraPosition, Vertex<3> lookAt, Vertex<3> up,
		double fovy, double aspect, int imageSize[2], Vertex<3> globalAmbient,
		Vertex<3> backgroundColor, double defaultRefractiveIndex,
		const std::vector<Object> &objects, const std::vector<Light> &lights,
		int maxRecursion, int oversample, int debugx, int debugy) {
		Image image;
		double viewSize[2];

		viewSize[1] = abs(cameraPosition - lookAt) * 2.0 * tan(fovy/2.0*M_PI/180.0);
		viewSize[0] = aspect * viewSize[1];

		image.width = imageSize[0];
		image.height = imageSize[1];
		image.bpp = 3;
		image.data.resize(image.width * image.height * image.bpp);

		Vertex<3> look = normalize(lookAt - cameraPosition);
		Vertex<3> left = normalize(crossProduct(up, look));

		if (debugx != -1 && debugy != -1) {
			int x = debugx;
			int y = debugy;

			gDebug = true;

			Vertex<3> target = normalize(lookAt + left * (imageSize[0] / 2 - x) / imageSize[0] * viewSize[0] + up * (imageSize[1] / 2 - y) / imageSize[1] * viewSize[1]);
			Vertex<3> color = traceRay(cameraPosition, target, objects, lights, maxRecursion, 1.0, globalAmbient, backgroundColor, defaultRefractiveIndex);

			gDebug = false;

			for (int i = 0; i < 3; i++) {
				if (color.v[i] > 1.0) color.v[i] = 1.0;
			}

			std::cout << "Final color for (" << x << ", " << y << ") = " << color << "\n";
		} else {
			for (int y = 0; y < imageSize[1]; y++) {
				for (int x = 0; x < imageSize[0]; x++) {
					const Vertex<3> target = normalize(
						look
						+ left * (imageSize[0] / 2 - x) * viewSize[0] / imageSize[0]
						+ up * (imageSize[1] / 2 - y) * viewSize[1] / imageSize[1]
					);
					Vertex<3> color = {0.0, 0.0, 0.0};

					if (oversample < 2) {
						color = traceRay(cameraPosition, target, objects, lights, maxRecursion, 1.0, globalAmbient, backgroundColor, defaultRefractiveIndex);
					} else {
						for (int i = 0; i < oversample; i++) {
							for (int j = 0; j < oversample; j++) {
								const Vertex<3> target = normalize(
									look
									+ left * (imageSize[0] / 2 - (x + (i - oversample / 2.0) / oversample)) * viewSize[0] / imageSize[0]
									+ up * (imageSize[1] / 2 - (y + (j - oversample / 2.0) / oversample)) * viewSize[1] / imageSize[1]
								);
								color = color + traceRay(cameraPosition, target, objects, lights, maxRecursion, 1.0, globalAmbient, backgroundColor, defaultRefractiveIndex);
							}
						}

						color = color / static_cast<double>(oversample * oversample);
					}

					for (int i = 0; i < 3; i++) {
						image.data[image.bpp * (image.width * (image.height - y - 1) + x) + i] = static_cast<unsigned char>(color.v[i] * 255);
					}

					// Progress-o-meter
					std::cout << '\r' << static_cast<int>(100.0 * (y * imageSize[0] + x) / (imageSize[0] * imageSize[1])) << '%' << std::flush;
				}
			}
		}

		std::cout << '\n';

		return image;
	}

	void testSphere(const Vertex<3> &origin, const Vertex<3> &direction, const Object &object, Vertex<3> &P, Vertex<3> &N, double &T, const Object *&O, bool &into) {
		const Vertex<3> &center = object.object.sphere.center;
		double radius = object.object.sphere.radius;
		Vertex<3> D = (origin - center);
		double a = dotProduct(direction, direction);
		double b = 2 * dotProduct(direction, D);
		double c = dotProduct(D, D) - radius * radius;
		double discriminant = b * b - 4 * a * c;
		double t = -1.0;

		if (discriminant > 0) {
			double sd = sqrt(discriminant);
			double t1 = (-b - sd) / (2 * a);
			double t2 = (-b + sd) / (2 * a);

			if (t1 >= RAY_EPSILON) {
				t = t1;
			} else if (t2 >= RAY_EPSILON) {
				t = t2;
			}
		} else if (discriminant == 0) {
			double t1 = -b / (2 * a);
			if (t1 >= 0) {
				t = t1;
			}
		}

		if (t >= 0 && (t < T || T < 0)) {
			T = t;
			P = origin + direction * t;
			if (dotProduct(direction, center - P) < 0.0) {
				N = (center - P) / radius;
				into = false;
			} else {
				N = (P - center) / radius;
				into = true;
			}
			O = &object;
		}
	}

	bool sameSide(const Vertex<3> &p1, const Vertex<3> &p2, const Vertex<3> &a, const Vertex<3> &b) {
		return dotProduct(crossProduct(b - p1, p1 - a), crossProduct(b - a, p2 - a)) >= 0;
	}

	bool inside(const Vertex<3> &t0, const Vertex<3> &t1, const Vertex<3> &t2, const Vertex<3> &p) {
		return sameSide(p, t0, t1, t2) && sameSide(p, t1, t0, t2) && sameSide(p, t2, t0, t1);
	}

	void testTriangle(const Vertex<3> &origin, const Vertex<3> &direction, const Object &object, Vertex<3> &P, Vertex<3> &N, double &T, const Object *&O, bool &into) {
		const Vertex<3> &vertex0 = object.object.triangle.vertex[0];
		const Vertex<3> &vertex1 = object.object.triangle.vertex[1];
		const Vertex<3> &vertex2 = object.object.triangle.vertex[2];

		const Vertex<3> Nplane = object.object.triangle.normal;

		double numerator = dotProduct(Nplane, origin) - dotProduct(Nplane, vertex0);
		double denominator = dotProduct(Nplane * -1.0, direction);

		if (denominator != 0.0) {
			double t1 = (numerator / denominator);
			const Vertex<3> Pthis = origin + direction * t1;

			if (inside(vertex0, vertex1, vertex2, Pthis) && t1 > RAY_EPSILON && (t1 < T || T < 0.0)) {
				T = t1;
				P = Pthis;
				O = &object;
				if (dotProduct(direction, Nplane) > 0.0) {
					N = Nplane * -1.0;
				} else {
					N = Nplane;
				}
				into = false;
			}
		}
	}

	Vertex<3> traceRayBack(const Vertex<3> &origin, const Vertex<3> &target, const std::vector<Object> &objects) {
		double T;
		Vertex<3> S;
		const Object *O = 0;
		Vertex<3> color = {1.0, 1.0, 1.0};
		bool into = false;
		
		for (size_t i = 0; i < objects.size(); i++) {
			T = -1.0;
			switch (objects[i].type) {
				case Object_Sphere:
					testSphere(origin, normalize(target - origin), objects[i], S, S, T, O, into);
					break;
				case Object_Triangle:
					testTriangle(origin, normalize(target - origin), objects[i], S, S, T, O, into);
					break;
			}
			if (T >= 0.0) {
				color = memberwiseProduct(color, objects[i].transmittance);
			}
		}

		return color;
	}

	Vertex<3> traceRay(const Vertex<3> &origin, const Vertex<3> &direction, const std::vector<Object> &objects, const std::vector<Light> &lights, int ttl, double refractiveIndex, const Vertex<3> &globalAmbient, const Vertex<3> &backgroundColor, double defaultRefractiveIndex) {
		Vertex<3> color = {0.0, 0.0, 0.0};
		double Tlast = -1.0;
		Vertex<3> Plast;
		Vertex<3> Nlast;
		const Object *Olast = 0;
		bool into = false;

		for (size_t i = 0; i < objects.size(); i++) {
			switch (objects[i].type) {
				case Object_Sphere:
					testSphere(origin, direction, objects[i], Plast, Nlast, Tlast, Olast, into);
					break;
				case Object_Triangle:
					testTriangle(origin, direction, objects[i], Plast, Nlast, Tlast, Olast, into);
					break;
			}
		}

		if (Tlast > 0.0) {
			if (gDebug) std::cout << "Hit " << ((Olast->type == Object_Sphere)?"sphere":"triangle") << " number " << (Olast - &*objects.begin()) << " at " << Plast << ".\n";

			const Vertex<3> R = direction - Nlast * dotProduct(direction * 2, Nlast);

			// Texture component
			Vertex<3> texture = {1.0, 1.0, 1.0};
			if (Olast->texture > 0) { // This part is certainly a hack, but it gets the job done
				if (Olast->type == Object_Sphere) {
					Vertex<3> coord = (Plast - Olast->object.sphere.center) / Olast->object.sphere.radius;
	
					double theta = atan2(coord.v[2], coord.v[0]);
					double phi = acos(coord.v[1]);
	
					bool t0 = int(2.0 * Olast->texture * theta / M_PI) % 2 != 0;
					bool t1 = int(2.0 * Olast->texture * phi / M_PI) % 2 != 0;
					if (t0 ^ t1) {
						texture.v[0] = texture.v[1] = texture.v[2] = 0.0;
					}
				} else if (Olast->type == Object_Triangle) {
					const int q = 1;
					const int r = 0;
					const int s = 2;
					const Vertex<3> d = Plast - Olast->object.triangle.vertex[q];
					double a = dotProduct(d, Olast->object.triangle.vertex[r] - Olast->object.triangle.vertex[q])
						/ dotProduct(Olast->object.triangle.vertex[r] - Olast->object.triangle.vertex[q], Olast->object.triangle.vertex[r] - Olast->object.triangle.vertex[q]);
					double b = dotProduct(d, Olast->object.triangle.vertex[s] - Olast->object.triangle.vertex[q])
						/ dotProduct(Olast->object.triangle.vertex[s] - Olast->object.triangle.vertex[q], Olast->object.triangle.vertex[s] - Olast->object.triangle.vertex[q]);

					bool t0 = int(a * Olast->texture) % 2 != 0;
					bool t1 = int(b * Olast->texture) % 2 != 0;

					if (t0 ^ t1) {
						texture.v[0] = texture.v[1] = texture.v[2] = 0.0;
					}
				}
			}
			
			for (size_t l = 0; l < lights.size(); l++) {
				const Light &light = lights[l];

				const Vertex<3> lightVector = normalize(light.position - Plast);

				const Vertex<3> back = traceRayBack(Plast, lights[l].position, objects);
				//if (dotProduct(back, back) > 0.0) back.v[0] = back.v[1] = back.v[2] = 1.0;

				// Diffuse component
				Vertex<3> diffuse = {0.0, 0.0, 0.0};
				const double c = dotProduct(Nlast, lightVector);
				if (c >= 0) {
					diffuse = memberwiseProduct(memberwiseProduct(Olast->diffuse, light.color) * c, texture);
					if (gDebug) std::cout << "DIFFUSE = " << diffuse << "\n";
				}

				// Specular component
				Vertex<3> specular = {0.0, 0.0, 0.0};
				const double spec = dotProduct(lightVector, R);

				if (spec > 0) {
					specular = memberwiseProduct(Olast->specular, light.color) * pow(spec, Olast->specularExponent);
					if (gDebug) std::cout << "SPECULAR = " << specular << "\n";
				}

				// Now combine all of the components
				color = color + specular + memberwiseProduct(diffuse, back);
			}
				
			if (ttl-- > 0) {
				if (gDebug) std::cout << "TTL = " << ttl << "\n";

				// Transmitted component
				if (dotProduct(Olast->transmittance, Olast->transmittance) > 0.0) {
					const double cosI = dotProduct(Nlast, direction);
					const double n = into?(refractiveIndex / Olast->refractiveIndex):(Olast->refractiveIndex / defaultRefractiveIndex);
					if (gDebug) {
						std::cout << "OR: " << Olast->refractiveIndex << "\n";
						std::cout << "R: " << refractiveIndex << "\n";
					}
					const double ri = into?Olast->refractiveIndex:defaultRefractiveIndex;
					const double c1 = -dotProduct(direction, Nlast);
					const double t = 1 - n * n * (1 - c1 * c1);

					if (t >= 0.0) {
						const double c2 = sqrt(1 - n * n * (1 - c1 * c1));
						const Vertex<3> T = direction * n + Nlast * (n * c1 - c2);

						if (gDebug) {
							std::cout << "TRANSMITTED: cosI = " << cosI << ", direction = " << direction << ", T = " << T << ", n = " << n << "\n";
						}
						const Vertex<3> transmitted = traceRay(Plast, T, objects, lights, ttl, ri, globalAmbient, backgroundColor, defaultRefractiveIndex);
						color = color + memberwiseProduct(transmitted, Olast->transmittance);
						if (gDebug) {
							std::cout << "TRANSMITTED: color = " << memberwiseProduct(transmitted, Olast->transmittance) << "\n";
						}
					}
				}

				// Reflected component
				if (dotProduct(Olast->reflective, Olast->reflective) > 0.0) {
					const Vertex<3> reflected = memberwiseProduct(traceRay(Plast, R, objects, lights, ttl, refractiveIndex, globalAmbient, backgroundColor, defaultRefractiveIndex), Olast->reflective);
					if (gDebug) std::cout << "REFLECTED = " << reflected << "\n";
					color = color + reflected;
				}

			}

			// Ambient component
			const Vertex<3> ambient = memberwiseProduct(memberwiseProduct(globalAmbient, Olast->ambient), texture);
			if (gDebug) std::cout << "AMBIENT = " << ambient << "\n";
			color = color + ambient;
		} else {
			color = backgroundColor;
		}

		// Clamp the color to reasonable values
		for (int i = 0; i < 3; i++) {
			if (color.v[i] > 1.0) color.v[i] = 1.0;
			if (color.v[i] < 0.0) color.v[i] = 0.0;
		}

		return color;
	}
}
